/*
 * Cresques Mapping Suite. Graphic Library for constructing mapping applications.
 * 
 * Copyright (C) 2004-5. 
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 * 
 * cresques@gmail.com
 */
package org.cresques.io;

import java.awt.Image;
import java.awt.Point;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.geom.Point2D;
import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;
import java.io.IOException;
import java.util.Vector;

import org.cresques.cts.ICoordTrans;
import org.cresques.cts.IProjection;
import org.cresques.filter.RasterBuf;
import org.cresques.io.data.Metadata;
import org.cresques.io.exceptions.NotSupportedExtensionException;
import org.cresques.px.Extent;

import es.gva.cit.jgdal.Gdal;
import es.gva.cit.jgdal.GdalBuffer;
import es.gva.cit.jgdal.GdalException;
import es.gva.cit.jgdal.GdalRasterBand;
import es.gva.cit.jgdal.GeoTransform;
/**
 * Soporte 'nativo' para ficheros desde GDAL.
 * Este conjunto de funcionalidades est tomado de manera casi literal
 * del soporte para ECW de ermapper.<br>
 * Probablemente esto debera formar parte del JNI que recubre a la
 * librera en C extraida de gdal.<br>
 * Lo pongo aqu a manera de ejemplo de como atacar un formato binario
 * desde Java.<br><br>   
 * @author Luis W. Sevilla.
 */

class GdalNative extends Gdal {
	static boolean 				WITH_OVERVIEWS = true;
	private String 				ext = "";
	/**
	 * Nombre corto del driver de gdal
	 */
	private String 				shortName = "";
	public 	GeoTransform 		trans = null;
	/**
	 * Contorno en coordenadas geogrficas. (y Extent del raster).
	 */
	public Contour 				bBoxRot = new Contour();
	/**
	 * Contorno en coordenadas geogrficas sin rotacin aplicada. Esto es util para poder
	 * calcular los pixeles necesarios que se van a leer del raster. Cuando el raster no tiene 
	 * rotacin coincide con esq. 
	 */
	public Contour				bBoxWithoutRot = new Contour();
	public int 					width = 0, height = 0;
	public double 				originX = 0D, originY = 0D;
	public String 				version = "";
	private int 				alpha = 0;
	protected int 				rBandNr = 1, gBandNr = 2, bBandNr = 3, aBandNr = 4;
	private int 				dataType = GDT_Byte;
	/**
	 * Metadatos leidos de la imagen
	 */
	private Metadata			metadata = null;
	private boolean 			georeferenced = true;

	
	// Polilinea con extent
	public class Contour extends Vector {
		final private static long serialVersionUID = -3370601314380922368L;
		public double minX = Double.MAX_VALUE, minY = Double.MAX_VALUE;
		public double maxX = -Double.MAX_VALUE, maxY = -Double.MAX_VALUE;
		public Contour() {
			super();
		}
		public void add(Point2D pt) {
			super.add(pt);
			if (pt.getX() > maxX) maxX = pt.getX();
			if (pt.getX() < minX) minX = pt.getX();
			if (pt.getY() > maxY) maxY = pt.getY();
			if (pt.getY() < minY) minY = pt.getY();
		}
	}
		
	public GdalNative(String fName) throws GdalException, IOException {
		super();
		init(fName);
	}
	
	/**
	 * <P>
	 * Calcula la bounding box en la que est metido el raster teniendo en cuenta
	 * el tamao de pixel y la rotacin. Esto lo hace con los valores de transformacin 
	 * leidos por gdal en el vector de 6 elementos adfGeoTransform donde cada elemento
	 * del vector represnta los siguientes valores
	 * </P>
	 * <UL>
	 * <LI>0-origen X</LI>
	 * <LI>1-tamao de pixel X</LI>
	 * <LI>2-shear en X</LI>
	 * <LI>3-origen Y</LI>
	 * <LI>4-shear en Y</LI>
	 * <LI>5-Tamao de pixel Y</LI>
	 * </UL>
	 * <P>
	 * Para el calculo de una esquina aplicamos la formula siguiente:<BR>
	 * PtoX = originX + pixelSizeX * x + shearX * y;<BR>
	 * PtoY = originY + shearY * x + pixelSizeY * y;<BR>
	 * Aplicandolo a las cuatro esquinas sustituimos en cada una de ellas por.
	 * </P>
	 * <UL> 
	 * <LI>Esquina superior izquierda: x = 0; y = 0;</LI>
	 * <LI>Esquina superior derecha: x = MaxX; y = 0;</LI>
	 * <LI>Esquina inferior izquierda: x = 0; y = MaxY;</LI>
	 * <LI>Esquina inferior derecha: x = MaxX; y = MaxY;</LI>
	 * </UL> 
	 * <P>
	 * quedandonos en los cuatro casos:
	 * </P>
	 * <UL> 
	 * <LI>Esquina superior izquierda: originX; originY;</LI>
	 * <LI>Esquina superior derecha: PtoX = originX + pixelSizeX * x; PtoY = originY + shearY * x;</LI>
	 * <LI>Esquina inferior izquierda:  PtoX = originX + shearX * y; PtoY = originY + pixelSizeY * y;</LI>
	 * <LI>Esquina inferior derecha: PtoX = originX + pixelSizeX * x + shearX * y; PtoY = originY + shearY * x + pixelSizeY * y;</LI>
	 * </UL>
	 * 
	 */
	private void boundingBoxFromGeoTransform(){
		double geoX = 0D, geoY = 0D;

		//Upper left corner
		bBoxRot.add(new Point2D.Double(trans.adfgeotransform[0], trans.adfgeotransform[3]));

		//Lower left corner
		geoX = trans.adfgeotransform[0] +  trans.adfgeotransform[2] * height;
		geoY = trans.adfgeotransform[3] +  trans.adfgeotransform[5] * height;
		bBoxRot.add(new Point2D.Double(geoX, geoY));
		
		//Upper right corner
		geoX = trans.adfgeotransform[0] + trans.adfgeotransform[1] * width;
		geoY = trans.adfgeotransform[3] + trans.adfgeotransform[4] * width;
		bBoxRot.add(new Point2D.Double(geoX, geoY));
		
		//Lower right corner
		geoX = trans.adfgeotransform[0] + trans.adfgeotransform[1] * width + trans.adfgeotransform[2] * height;
		geoY = trans.adfgeotransform[3] + trans.adfgeotransform[4] * width + trans.adfgeotransform[5] * height;
		bBoxRot.add(new Point2D.Double(geoX, geoY));
		
		//TODO: OJO! con coordenadas geogrficas
	}
	
	/**
	 * Calcula la bounding box en la que est metido el raster teniendo en cuenta
	 * el tamao de pixel y la rotacin. 
	 */
	private void boundingBoxWithoutRotation(){
		double ox = trans.adfgeotransform[0];
		double oy = trans.adfgeotransform[3];
		double resx = trans.adfgeotransform[1];
		double resy = trans.adfgeotransform[5];
				
		bBoxWithoutRot.add(new Point2D.Double(ox, oy));
		bBoxWithoutRot.add(new Point2D.Double(ox + resx * width, oy));
		bBoxWithoutRot.add(new Point2D.Double(ox, oy + resy * height));
		bBoxWithoutRot.add(new Point2D.Double(ox + resx * width, oy + resy * height));

		//TODO: OJO! con coordenadas geogrficas
	}
	
	private void init(String fName) throws GdalException, IOException {
		open(fName,GA_ReadOnly);
		ext = fName.toLowerCase().substring(fName.lastIndexOf('.')+1);
		if (ext.compareTo("tif") == 0)
			WITH_OVERVIEWS = false;
		width = getRasterXSize();
		height = getRasterYSize();
		setDataType(this.getRasterBand(1).getRasterDataType());
		shortName = getDriverShortName();
		metadata = new Metadata(getMetadata());
		
		//Asignamos la interpretacin de color leida por gdal a cada banda. Esto nos sirve
		//para saber que banda de la imagen va asignada a cada banda de visualizacin (ARGB)
		metadata.initColorInterpretation(getRasterCount());
		metadata.initNoDataByBand(getRasterCount());
	    for(int i = 0; i < getRasterCount(); i++){
	    	GdalRasterBand rb = getRasterBand(i + 1);
	    	String colorInt = getColorInterpretationName(rb.getRasterColorInterpretation());
	    	metadata.setNoDataValue(i, rb.getRasterNoDataValue());
	    	metadata.setColorInterpValue(i, colorInt);
	    	if(colorInt.equals("Red"))
	    		rBandNr = i + 1;
	    	if(colorInt.equals("Green"))
	    		gBandNr = i + 1;	
	    	if(colorInt.equals("Blue"))
	    		bBandNr = i + 1;		
	    	if(colorInt.equals("Alpha"))
	    		aBandNr = i + 1;
	    }
	    
		double ox=0D, oy=0D, resx=0D, resy=0D;
		try{
			trans = getGeoTransform();
			
			boundingBoxWithoutRotation();
			boundingBoxFromGeoTransform();
						
			this.georeferenced = true;
		}catch(GdalException exc){
			bBoxRot.add(new Point2D.Double(0, 0));
			bBoxRot.add(new Point2D.Double(width, 0));
			bBoxRot.add(new Point2D.Double(0, height));
			bBoxRot.add(new Point2D.Double(width, height));
			bBoxWithoutRot = bBoxRot;
			this.georeferenced = false;
		}
	}
	
	public void setAlpha(int a) { alpha = a; }
	
	public void setDataType(int dt) { dataType = dt; }
	public int getDataType() { return dataType; }
	
	double lastReadLine = -1;
	int currentFullWidth = -1;
	int currentFullHeight = -1;
	int currentViewWidth = -1;
	int currentViewHeight = -1;
	double currentViewX = 0D;
	double currentViewY = 0D;
	double viewportScaleX = 0D;
	double viewportScaleY = 0D;
	double wcWidth = 0D;
	double stepX = 0D;
	double stepY = 0D;
	int currentOverview = -1;
	
	protected GdalRasterBand bandR = null, bandG = null, bandB = null, bandA = null;
	
	private boolean[] orientation;
	
	/**
	 * Devuelve la banda actualmente en uso para el color especificado.
	 * @param color 0=Rojo, 1=Green, 2=Blue.
	 * @return
	 */
	public GdalRasterBand getCurrentBand(int color) {
		if (color == 0) 
			return bandR;
		else if (color == 1) 
			return bandG;
		return bandB;
	}
	
	//Supone rasters no girados
	public Point2D worldToRaster(Point2D pt) {
		double x = (((double) currentFullWidth) / (bBoxWithoutRot.maxX - bBoxWithoutRot.minX)) * (pt.getX() - bBoxWithoutRot.minX);
		double y = (((double) currentFullHeight) / (bBoxWithoutRot.maxY - bBoxWithoutRot.minY)) * (bBoxWithoutRot.maxY - pt.getY());
		Point2D ptRes = new Point2D.Double(x, y);
		return ptRes;
	}
	
	/**
	 * Si el tamao de pixel en X es menor que 0 entonces la imagen se orienta al contrario en X por lo que en los zooms
	 * habr que invertir la peticin de la parte derecha a la izquierda y viceversa. Esto lo detectamos con la 
	 * variable orientation , si orientation[0] es false entonces el punto inicial del zoom lo invertimos de la 
	 * siguiente forma:
	 * Nuevo_punto_inicialX = (Ancho_total_raster - punto_inicial_del_zoomX) - Ancho_de_peticin
	 *
	 * Si el tamao de pixel en Y es mayor que 0 entonces la imagen se orienta al contrario en Y por 
	 * lo que en los zooms habr que invertir la peticin de abajo a arriba y viceversa. Esto lo detectamos con la 
	 * variable orientation , si orientation[1] es true entonces el punto inicial del zoom lo invertimos de la 
	 * siguiente forma:
	 * Nuevo_punto_inicialY = (Alto_total_raster - punto_inicial_del_zoomY) - Alto_de_peticin
	 * 
	 * @param dWorldTLX
	 * @param dWorldTLY
	 * @param dWorldBRX
	 * @param dWorldBRY
	 * @param nWidth
	 * @param nHeight
	 * @param orientation array de dos elementos que representa la orientacin de la peticin en
	 * X e Y. El primer elemento representa el signo de pixelSize en X, true si es positivo y false
	 * si es negativo. El segundo elemento representa el signo de pixelSize en Y
	 * @return
	 */
	public int setView(double dWorldTLX, double dWorldTLY,
            double dWorldBRX, double dWorldBRY,
            int nWidth, int nHeight, boolean[] orientation) {
		int err = 0;
		this.orientation = orientation;
		currentFullWidth = width;
		currentFullHeight = height;
		Point2D tl = worldToRaster(new Point2D.Double(dWorldTLX, dWorldTLY));
		Point2D br = worldToRaster(new Point2D.Double(dWorldBRX, dWorldBRY));
		// Calcula cual es la primera lnea a leer;
		currentViewWidth = nWidth;
		currentViewHeight = nHeight;
		wcWidth = Math.abs(br.getX() - tl.getX());
		
		if(!orientation[0]) //Invierte la orientacin en X
			currentViewX = (width - tl.getX()) - (br.getX()-tl.getX());
		else
			currentViewX = tl.getX();
		
		viewportScaleX = (double) currentViewWidth/(br.getX()-tl.getX());
		viewportScaleY = (double) currentViewHeight/(br.getY()-tl.getY());
		stepX = 1D/viewportScaleX;
		stepY = 1D/viewportScaleY;

		if(orientation[1])//Invierte la orientacin en Y
			lastReadLine = (height - tl.getY()) - (br.getY()-tl.getY());
		else
			lastReadLine = tl.getY();
		
		try {
			// calcula el overview a usar
			bandR = getRasterBand(1);
			currentOverview = -1;
			if (WITH_OVERVIEWS && bandR.getOverviewCount() > 0) {
				GdalRasterBand ovb = null;
				for (int i=bandR.getOverviewCount()-1; i>0; i--) {              
					ovb = bandR.getOverview(i);
					if (ovb.getRasterBandXSize()>getRasterXSize()*viewportScaleX) {
						currentOverview = i;
	            		viewportScaleX *= ((double) width/(double) ovb.getRasterBandXSize());
	            		viewportScaleY *= ((double) height/(double) ovb.getRasterBandYSize());
	            		stepX = 1D/viewportScaleX;
	            		stepY = 1D/viewportScaleY;
	            		currentFullWidth = ovb.getRasterBandXSize();
	            		currentFullHeight = ovb.getRasterBandYSize();
	            		tl = worldToRaster(new Point2D.Double(dWorldTLX, dWorldTLY));
	            		if(!orientation[0])//Invierte la orientacin en X
	            			currentViewX = (width - tl.getX()) - (br.getX()-tl.getX());
	            		else
	            			currentViewX = tl.getX();
	            		if(orientation[1])//Invierte la orientacin en Y
	            			lastReadLine = (height - tl.getY()) - (br.getY()-tl.getY());
	            		else
	            			lastReadLine = tl.getY();
	            		break;
					}
				}
			}
	
			// Selecciona las bandas y los overviews necesarios
			bandR = getRasterBand(rBandNr);
			setDataType(bandR.getRasterDataType());
			
			if (this.getRasterCount() > 1) {
				bandG = getRasterBand(gBandNr);
				bandB = getRasterBand(bBandNr);				
				if(metadata.isAlphaBand())
					bandA = getRasterBand(aBandNr); 
			}
			if (currentOverview > 0) {
				bandR = bandR.getOverview(currentOverview);
    			if (this.getRasterCount() > 1) {
					bandG = bandG.getOverview(currentOverview);
					bandB = bandB.getOverview(currentOverview);
					if(metadata.isAlphaBand())
						bandA = bandA.getOverview(currentOverview);
				}
			}

		} catch (GdalException e) {
			e.printStackTrace();
		}
		return err;
	}
	
	int lastY = -1;
	
	public void readLine(int[][] line) throws GdalException {
        int w = (int) (Math.ceil(((double)currentViewWidth)*stepX) + 1);
        int x = (int) Math.ceil(currentViewX);
        int y = (int) Math.ceil(lastReadLine);
        GdalBuffer r = null, g = null, b = null, p = null;
        GdalBuffer a = new GdalBuffer();
        
        //if (alpha > 0) a = alpha << 24;
        if (x+w > bandR.getRasterBandXSize()) 
        	w = bandR.getRasterBandXSize()-x;
        
        if(bandR.getRasterColorTable() != null){
        	p = bandR.readRasterWithPalette(x, y, w, 1, w, 1, dataType);
        	a.buffByte = p.buffAPalette;
        	r = new GdalBuffer();
        	r.buffByte = p.buffRPalette;
        	g = new GdalBuffer();
        	g.buffByte = p.buffGPalette;
        	b = new GdalBuffer();
        	b.buffByte = p.buffBPalette;
        }else{
        	a.buffByte = new byte[w];
			r = bandR.readRaster(x, y, w, 1, w, 1, dataType);
			if (bandG != null)
	    		g = bandG.readRaster(x, y, w, 1, w, 1, dataType);
			if (bandB != null)
	    		b = bandB.readRaster(x, y, w, 1, w, 1, dataType);
        }
          
		lastReadLine += stepY;
        
  		int i=0;
  		double j = 0D;
  		double initOffset =  Math.abs(currentViewX - ((int)currentViewX));
  		
  		if (dataType == GDT_CInt16 || dataType == GDT_Int16  || dataType == GDT_UInt16){
  			if (g == null){ // Sibgle Band (Typical DEM)
  				for (int k=0; k<4; k++){
  					for (i=0, j = initOffset; i<currentViewWidth && j<r.getSize(); i++, j+=stepX) {
  						if(k<3)
  							line[i][k] = (r.buffShort[(int) j] & 0xffff);
  						else
  							line[i][3] = 0xff;
  					}
	      		}
  			}else { // Multiband
  				//System.err.println("readLine(): Raster 16bits multibanda");
  				GdalBuffer [] bands = {r,g,b};
  				for (int k=0; k<4; k++){
		      		for (i=0, j = initOffset; i<currentViewWidth && j<r.getSize(); i++, j+=stepX){
		      			if(k<3)
		      				line[i][k] = (bands[k].buffShort[(int) j] & 0xffff);
		      			else
		      				line[i][3] = 0xff;
		      		}
  				}
  			}
  		}else if(dataType == GDT_Float32){
  			GdalBuffer [] bands = {r,g,b};
			for (int k=0; k<4; k++){
	      		for (i=0, j = initOffset; i<currentViewWidth && j<r.getSize(); i++, j+=stepX){
	      			if(k < 3)
	      				line[i][k] = (int)bands[0].buffFloat[(int) j];
	      			else
	      				line[i][3] = 0xff;
	      		}
			}
  		}
  		
		return;
	}
	
	//int liney = 0;
	int readLineRGBA(int [] line) throws GdalException {
		int err = 0;
		
        int w = (int) (Math.ceil(((double)currentViewWidth)*stepX) + 1);
        int x = (int) currentViewX;
        int y = (int) lastReadLine;
        GdalBuffer r = null, g = null, b = null, p = null;
        GdalBuffer a = new GdalBuffer();
        
        while(y >= bandR.getRasterBandYSize())
        	y--;
        
        //if (alpha > 0) a = alpha << 24;
        if (x+w > bandR.getRasterBandXSize()) 
        	w = bandR.getRasterBandXSize()-x;
                       
        if(bandR.getRasterColorTable() != null){
        	p = bandR.readRasterWithPalette(x, y, w, 1, w, 1, dataType);
        	a.buffByte = p.buffAPalette;
        	r = new GdalBuffer();
        	r.buffByte = p.buffRPalette;
        	g = new GdalBuffer();
        	g.buffByte = p.buffGPalette;
        	b = new GdalBuffer();
        	b.buffByte = p.buffBPalette;
        }else{			
        	r = bandR.readRaster(x, y, w, 1, w, 1, dataType);
			if (bandG != null)
	    		g = bandG.readRaster(x, y, w, 1, w, 1, dataType);
			if (bandB != null)
	    		b = bandB.readRaster(x, y, w, 1, w, 1, dataType);
        	
        	if(metadata.isAlphaBand()){
        	//if(getRasterCount() == 4 && shortName.equals("PNG")){
        		a = bandA.readRaster(x, y, w, 1, w, 1, GDT_Byte);	
        	}else{
    			a.buffByte = new byte[w];
    			for (int i = 0;i < w;i++)
    				a.buffByte[i] = (byte)255;
        	}
        }
       
        lastReadLine += stepY;
  	
  		int i=0;
  		double j =  Math.abs(currentViewX - ((int)currentViewX));
		int alpha = (this.alpha & 0xff) << 24;

		if(orientation[0]){ //Pixel size en X positivo
	  		if (dataType == GDT_Byte){
		  		if (g != null)
		      		for (i=0; i<currentViewWidth && j<r.getSize(); i++, j+=stepX) {
		      			int jInt = (int)(j);
		      			line[i] = (alpha & ((a.buffByte[jInt])& 0xff) << 24) + ((r.buffByte[jInt] & 0xff) << 16) + ((g.buffByte[jInt] & 0xff) << 8) + (b.buffByte[jInt] & 0xff);
		      		}
		      	else
		      		for (i=0; i<currentViewWidth && j<r.getSize(); i++, j+=stepX) {
		      			int jInt = (int)(j);
		      			line[i] = (alpha & ((a.buffByte[jInt])& 0xff) << 24) + ((r.buffByte[jInt] & 0xff) << 16) + ((r.buffByte[jInt] & 0xff) << 8) + (r.buffByte[jInt] & 0xff);
		      		}
	  		}else if (dataType == GDT_CInt16 || dataType == GDT_Int16  || dataType == GDT_UInt16){
	  			if (g == null) // Sibgle Band (Typical DEM)
	  		  		for (i=0; i<currentViewWidth && j<r.getSize(); i++, j+=stepX) {
	  		  			int jInt = (int)(j);
		      			line[i] = (alpha & ((a.buffByte[jInt])& 0xff) << 24) + r.buffShort[jInt];
		      		}
	  			else { // Multiband - Raster 16bits multibanda
		      		for (i=0; i<currentViewWidth && j<r.getSize(); i++, j+=stepX) {
		      			int jInt = (int)(j);
		      			line[i] = (alpha & ((a.buffByte[jInt])& 0xff) << 24) | (((r.buffShort[jInt] & 0xfff0) << 12) & 0xff0000 ) | 
										  						 (((g.buffShort[jInt] & 0xfff0) << 4 ) & 0xff00 ) |
										  						 (((b.buffShort[jInt] & 0xfff0) >> 4 ) & 0xff );
		      		}
	  			}
	  		}
		}else{ //Pixel size en X negativo
			if (dataType == GDT_Byte){
		  		if (g != null)
		      		for (i=currentViewWidth - 1; i>=0 && j<r.getSize(); i--, j+=stepX) {
		      			int jInt = (int)(j);
		      			line[i] = (alpha & ((a.buffByte[jInt])& 0xff) << 24) + ((r.buffByte[jInt] & 0xff) << 16) + ((g.buffByte[jInt] & 0xff) << 8) + (b.buffByte[jInt] & 0xff);
		      		}
		      	else
		      		for (i=currentViewWidth - 1; i>=0 && j<r.getSize(); i--, j+=stepX) {
		      			int jInt = (int)(j);
		      			line[i] = (alpha & ((a.buffByte[jInt])& 0xff) << 24) + ((r.buffByte[jInt] & 0xff) << 16) + ((r.buffByte[jInt] & 0xff) << 8) + (r.buffByte[jInt] & 0xff);
		      		}
	  		}else if (dataType == GDT_CInt16 || dataType == GDT_Int16  || dataType == GDT_UInt16){
	  			if (g == null) // Sibgle Band (Typical DEM)
	  				for (i=currentViewWidth - 1; i>=0 && j<r.getSize(); i--, j+=stepX) {
	  		  			int jInt = (int)(j);
		      			line[i] = (alpha & ((a.buffByte[jInt])& 0xff) << 24) + r.buffShort[jInt];
		      		}
	  			else { // Multiband - Raster 16bits multibanda;
	  				for (i=currentViewWidth - 1; i>=0 && j<r.getSize(); i--, j+=stepX) {
		      			int jInt = (int)(j);
		      			line[i] = (alpha & ((a.buffByte[jInt])& 0xff) << 24) | (((r.buffShort[jInt] & 0xfff0) << 12) & 0xff0000 ) | 
										  						 (((g.buffShort[jInt] & 0xfff0) << 4 ) & 0xff00 ) |
										  						 (((b.buffShort[jInt] & 0xfff0) >> 4 ) & 0xff );
		      		}
	  			}
	  		}
			
		}

		return err;
	}
		
	/**
	 * Lee una franja de la imagen.
	 * @param bandH Altura de la franja
	 * @param bufH	Altura del buffer
	 * @param buf	Buffer con la franja (retorno)
	 * @return
	 * @throws GdalException
	 */
	public int readBandRGBA(int bandH, int bufH, int [] buf) throws GdalException {
		int err = 0;
        int w = (int)(((double)currentViewWidth)*stepX);
        int x = (int)(((double)currentViewX)*stepX);
        int y = (int) lastReadLine;
        int h = (int) (((double)bandH)*stepX);
        System.out.println("Leyendo "+y);
        GdalBuffer r = null, g = null, b = null, p = null;
        GdalBuffer a = new GdalBuffer();
        
        if (x+w > bandR.getRasterBandXSize()) 
        	w = bandR.getRasterBandXSize()-x;
        
        if(bandR.getRasterColorTable() != null){
        	p = bandR.readRasterWithPalette(x, y, w, h, w, h, GDT_Byte);
        	a.buffByte = p.buffAPalette;
        	r = new GdalBuffer();
        	r.buffByte = p.buffRPalette;
        	g = new GdalBuffer();
        	g.buffByte = p.buffGPalette;
        	b = new GdalBuffer();
        	b.buffByte = p.buffBPalette;
        }else{
        	r = bandR.readRaster(x, y, w, h, w, h, dataType);
			if (bandG != null)
	    		g = bandG.readRaster(x, y, w, h, w, h, dataType);
			if (bandB != null)
	    		b = bandB.readRaster(x, y, w, h, w, h, dataType);
			
        	if(metadata.isAlphaBand()){
        	//if(getRasterCount() == 4 && shortName.equals("PNG")){
        		a = bandA.readRaster(x, y, w, h, w, h, GDT_Byte);
        	}else{
    			a.buffByte = new byte[w];
    			for (int i = 0;i < w*h;i++)
    				a.buffByte[i] = (byte)255;
        	}
        }
        
        lastReadLine += ((double)bandH)*stepY;
  		
  		// TODO Acabar de implementarlo
  		float k=0F;
		int alpha = (this.alpha & 0xff) << 24;
  		for (int j=0, t=0; j<bandH; j++) {
  			k = j*w; t=j*currentViewWidth;
  			for (int i=0; i<currentViewWidth && k<r.getSize(); i++, k+=stepX) {
  				buf[t+i] = (alpha & ((a.buffByte[(int)j])& 0xff) << 24) + ((r.buffByte[(int) k]) << 16) + ((g.buffByte[(int) k]) << 8) + b.buffByte[(int) k];
  			}
  		}
		
		return err;
		
	}

	/* (non-Javadoc)
	 * @see org.cresques.io.GeoRasterFile#getData(int, int, int)
	 */
	public Object[] getData(int x, int y) {
		try {
			Object[] data = new Object[getRasterCount()];
			for(int i = 0; i < getRasterCount(); i++){
				GdalRasterBand rb = getRasterBand(i + 1);
				GdalBuffer r = rb.readRaster(x, y, 1, 1, 1, 1, dataType);
				switch(dataType){
				case 0:	break;									//Sin tipo
				case 1:	data[i] = new Integer(r.buffByte[0]); 	//Buffer byte (8)
						break;
				case 2:											//Buffer short (16)
				case 3:	data[i] = new Integer(r.buffShort[0]);	//Buffer short (16)
						break;
				case 4:											//Buffer int (32)
				case 5: data[i] = new Integer(r.buffInt[0]);	//Buffer int (32)
						break;
				case 6:	data[i] = new Float(r.buffFloat[0]);	//Buffer float (32)
						break;
				case 7:	data[i] = new Double(r.buffDouble[0]);	//Buffer double (64)
						break;
				}
			}
			return data;
		} catch (GdalException e) {
			return null;
		}
	}
	
	void pintaInfo() {
		try {
			//System.out.println("Origin = "+originX+","+originY);
			//System.out.println("Origin = "+this.);
			System.out.println("GeoTransform:");
			GeoTransform trans = getGeoTransform();
			for (int i=0; i<6; i++)
				System.out.println("  param["+i+"]="+trans.adfgeotransform[i]);
			System.out.println("Metadata:");
			String [] metadata = getMetadata();
			for (int i=0; i<metadata.length; i++) {
				System.out.println(metadata[i]);
			}
		} catch (GdalException e) {
			
		}
		
	}
	
	void pintaPaleta() {
	}
	
	public int getBlockSize(){
		return this.getBlockSize();
	}

	/**
	 * Obtiene el objeto que contiene los metadatos
	 */
	public Metadata getMetadataJavaObject() {
		return metadata;
	}

	/**
	 * Obtiene el flag que dice si la imagen est o no georreferenciada
	 * @return true si est georreferenciada y false si no lo est.
	 */
	public boolean isGeoreferenced() {
		return georeferenced;
	}
}

/**
 * @author Luis W. Sevilla
 */
public class GdalFile extends GeoRasterFile {
	public final static int 	BAND_HEIGHT = 64;
	protected GdalNative 		file = null;

	private Extent v = null;
	
	public GdalFile(IProjection proj, String fName)throws NotSupportedExtensionException{
		super(proj, fName);
		extent = new Extent();
		try {
			file = new GdalNative(fName);
			load();
			readGeoInfo(fName);
			bandCount = file.getRasterCount(); 
			if ( bandCount > 2) {
				setBand(RED_BAND,   0);
				setBand(GREEN_BAND, 1);
				setBand(BLUE_BAND,  2);
			} else
				setBand(RED_BAND|GREEN_BAND|BLUE_BAND, 0);
		} catch (GdalException e) {
			throw new NotSupportedExtensionException("Extension not supported");
		} catch(Exception e){
		  	System.out.println("Error en GdalOpen");
		  	e.printStackTrace();
		  	file = null;
		}
		
		switch(file.getDataType()){
		case 1:setDataType(DataBuffer.TYPE_BYTE);break;//GDT_BYTE 
		case 2://GDT_UInt16 
		case 3:setDataType(DataBuffer.TYPE_SHORT);break;//GDT_Int16	
		case 4://GDT_UInt32
		case 5:setDataType(DataBuffer.TYPE_INT);break;//GDT_Int32
		case 6:setDataType(DataBuffer.TYPE_FLOAT);break;//GDT_Float32
		case 7:setDataType(DataBuffer.TYPE_DOUBLE);break;//GDT_Float64
		case 8:setDataType(DataBuffer.TYPE_UNDEFINED);break;//GDT_CInt16
		case 9:setDataType(DataBuffer.TYPE_UNDEFINED);break;//GDT_CInt32
		case 10:setDataType(DataBuffer.TYPE_UNDEFINED);break;//GDT_CFloat32
		case 11:setDataType(DataBuffer.TYPE_UNDEFINED);break;//GDT_CFloat64
		}
		
	}
	
	/**
	 * Obtenemos o calculamos el extent de la imagen.
	 */
	public GeoFile load() {
		extent = new Extent(file.bBoxRot.minX, file.bBoxRot.minY, file.bBoxRot.maxX, file.bBoxRot.maxY);
		requestExtent = new Extent(file.bBoxWithoutRot.minX, file.bBoxWithoutRot.minY, file.bBoxWithoutRot.maxX, file.bBoxWithoutRot.maxY);
		return this;
	}
	
	/**
	 * Cierra el fichero de imagen
	 */
	public void close() {
		try {
			if(file != null){
				file.close();
				file = null;
			}
		} catch (GdalException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	/**
	 * Asigna a cada banda R,G o B una banda de la imagen
	 */
	public void setBand(int flag, int bandNr) {
		super.setBand(flag, bandNr);
		if ((flag & GeoRasterFile.RED_BAND) == GeoRasterFile.RED_BAND) file.rBandNr = bandNr+1;
		if ((flag & GeoRasterFile.GREEN_BAND) == GeoRasterFile.GREEN_BAND) file.gBandNr = bandNr+1;
		if ((flag & GeoRasterFile.BLUE_BAND) == GeoRasterFile.BLUE_BAND) file.bBandNr = bandNr+1;
	}
	
	/**
	 * Asigna el extent de la vista actual. existe un fichero .rmf debemos hacer una transformacin
	 * de la vista asignada ya que la peticin viene en coordenadas del fichero .rmf y la vista (v)
	 * ha de estar en coordenadas del fichero.
	 */
	public void setView(Extent e) { 
		if(rmfExists){
												
			Point2D.Double petInit = null, petEnd = null;
			try{
				petInit = new Point2D.Double(e.minX(),  e.maxY());
				petEnd = new Point2D.Double(e.maxX(), e.minY());
				transformRMF.inverseTransform(petInit, petInit);
				transformRMF.inverseTransform(petEnd, petEnd);
				transformTFW.transform(petInit, petInit);
				transformTFW.transform(petEnd, petEnd);
			}catch(NoninvertibleTransformException ex){}
			double h = file.bBoxWithoutRot.maxY - file.bBoxWithoutRot.minY;
			if(!file.isGeoreferenced())
				v = new Extent(	petInit.getX(), h - petInit.getY(), petEnd.getX(), h - petEnd.getY()); 
			else
				v = new Extent(	petInit.getX(), petInit.getY(), petEnd.getX(), petEnd.getY());
			
		}else
			v = new Extent(e.minX(), e.minY(), e.maxX(), e.maxY());	
	}
	
	 /**
	 * Calcula la transformacin que se produce sobre la vista cuando la imagen tiene un fichero .rmf
	 * asociado. En Gdal el origen de coordenadas en Y es el valor mnimo y crece hasta el mximo. De la
	 * misma forma calcula la matriz de transformacin de la cabecera del fichero o del world file asociado
	 * @param originX Origen de la imagen en la coordenada X
	 * @param originY Origen de la imagen en la coordenada Y
	 */
	public void setExtentTransform(double originX, double originY, double psX, double psY) {		
		transformRMF.setToTranslation(originX, originY);
		transformRMF.scale(psX, psY);
		
		if(file.trans != null){	
			transformTFW.setToTranslation(file.trans.adfgeotransform[0], file.trans.adfgeotransform[3]);
			transformTFW.scale(file.trans.adfgeotransform[1], file.trans.adfgeotransform[5]);
		}
	}
		
	/**
	 * Obtiene extent de la vista actual
	 */
	public Extent getView() { 
		return v; 
	}
	
	/**
	 * Obtiene la anchura del fichero
	 */
	public int getWidth() {	
		return file.width; 
	}
	
	/**
	 * Obtiene la altura del fichero
	 */
	public int getHeight() { 
		return file.height;
	}

	/* (non-Javadoc)
	 * @see org.cresques.io.GeoRasterFile#reProject(org.cresques.cts.ICoordTrans)
	 */
	public void reProject(ICoordTrans rp) {
		// TODO Auto-generated method stub
	}
	
	/**
	 * Obtiene la orientacin de la imagen a partir del signo del tamao de pixel para poder
	 * asignarlo en el setView. Esto es util para poder conocer como debe leerse la image, 
	 * de abajo a arriba, de arriba a abajo, de izquierda a derecha o de derecha a izquierda. 
	 * La posicin habitual es la que el pixel size en X es positivo y en Y negativo leyendose 
	 * en este caso las X de menor a mayor y las Y de mayor a menor. Los casos posibles son:
	 * <UL>
	 * <LI><B>X > 0; Y < 0;</B> {true, false}</LI>
	 * <LI><B>X > 0; Y > 0;</B> {true, true}</LI>
	 * <LI><B>X < 0; Y > 0;</B> {false, true}</LI>
	 * <LI><B>X < 0; Y < 0;</B> {false, false}</LI>
	 * </UL>
	 *  
	 * @return
	 */
	private boolean[] getOrientation(){
		boolean[] orientation = {true, false};
		if(!rmfExists){
			if(file.trans != null && file.trans.adfgeotransform != null && file.trans.adfgeotransform[5] > 0)
				orientation[1] = true;
			if(file.trans != null && file.trans.adfgeotransform != null && file.trans.adfgeotransform[1] < 0)
				orientation[0] = false;
		}else{
			if(rmfTransform.getScaleY() > 0)
				orientation[1] = true;
			if(rmfTransform.getScaleX() < 0)
				orientation[0] = false;
		}
		return orientation;
	}
	
	/* (non-Javadoc)
	 * @see org.cresques.io.GeoRasterFile#updateImage(int, int, org.cresques.cts.ICoordTrans)
	 */
	public Image updateImage(int width, int height, ICoordTrans rp) {
		int line, pRGBArray[] = null;
		Image image = null;
			
		if (mustVerifySize()) {
			// Work out the correct aspect for the setView call.
			double dFileAspect = (double)v.width()/(double)v.height();
			double dWindowAspect = (double)width /(double)height;
	
			if (dFileAspect > dWindowAspect) {
			  height =(int)((double)width/dFileAspect);
			} else {
			  width = (int)((double)height*dFileAspect);
			}
		}
						
		// Set the view				
		file.setView(v.minX(), v.maxY(), v.maxX(), v.minY(),
			width, height, getOrientation());
		
		if(width<=0)width=1;
		if(height<=0)height=1;
		
		image = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
		//image = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
		pRGBArray = new int[width/**BAND_HEIGHT*/];
		try {
			//int nLin = height % BAND_HEIGHT;
			file.setAlpha(getAlpha());
			setBand(RED_BAND,   rBandNr);
			setBand(GREEN_BAND, gBandNr);
			setBand(BLUE_BAND,  bBandNr);
			for (line=0; line < height; line++) { //+=BAND_HEIGHT) {
				//int bandH = Math.min(BAND_HEIGHT, height-line);
				//file.readBandRGBA(bandH, BAND_HEIGHT, pRGBArray);
				file.readLineRGBA(pRGBArray);
				setRGBLine((BufferedImage) image, 0, line, width, 1/*bandH*/, pRGBArray, 0, width);
			}
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		return image;
	}
	
	public RasterBuf getRaster(int width, int height, ICoordTrans rp) {
		int line;
		RasterBuf raster = null;
			
		if(mustVerifySize()){
			// Work out the correct aspect for the setView call.
			double dFileAspect = (double)v.width()/(double)v.height();
			double dWindowAspect = (double)width /(double)height;
	
			if (dFileAspect > dWindowAspect) {
			  height =(int)((double)width/dFileAspect);
			} else {
			  width = (int)((double)height*dFileAspect);
			}
		}
		
		// Set the view
		boolean[] orientation = getOrientation();
		file.setView(v.minX(), v.maxY(), v.maxX(), v.minY(),
			width, height, orientation);
		
		raster = new RasterBuf(DataBuffer.TYPE_INT, width, height, 4, new Point(0,0));
		try {

			file.setAlpha(getAlpha());
			setBand(RED_BAND,   rBandNr);
			setBand(GREEN_BAND, gBandNr);
			setBand(BLUE_BAND,  bBandNr);
			for (line=0; line < height; line++) { //+=BAND_HEIGHT) {
				file.readLine(raster.getLineInt(line));
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		return raster;
	}
	
	/**
	 * Asigna al objeto Image los valores con los dato de la imagen contenidos en el 
	 * vector de enteros.
	 * @param image	imagen con los datos actuales
	 * @param startX	inicio de la posicin en X dentro de la imagen
	 * @param startY	inicio de la posicin en X dentro de la imagen
	 * @param w	Ancho de la imagen
	 * @param h	Alto de la imagen
	 * @param rgbArray	vector que contiene la banda que se va a sustituir
	 * @param offset	desplazamiento
	 * @param scansize	tamao de imagen recorrida por cada p
	 */
	protected void setRGBLine(BufferedImage image, int startX, int startY, int w, int h, int[] rgbArray, 
			 int offset, int scansize) {
		image.setRGB(startX, startY, w, h, rgbArray, offset, scansize);
	}
	
	/**
	 * Asigna al objeto Image la mezcla entre los valores que ya tiene y los valores 
	 * con los dato de la imagen contenidos en el vector de enteros. De los valores RGB
	 * que ya contiene se mantienen las bandas que no coinciden con el valor de flags. La
	 * banda correspondiente a flags es sustituida por los datos del vector.
	 * @param image	imagen con los datos actuales
	 * @param startX	inicio de la posicin en X dentro de la imagen
	 * @param startY	inicio de la posicin en X dentro de la imagen
	 * @param w	Ancho de la imagen
	 * @param h	Alto de la imagen
	 * @param rgbArray	vector que contiene la banda que se va a sustituir
	 * @param offset	desplazamiento
	 * @param scansize	tamao de imagen recorrida por cada paso
	 * @param flags	banda que se va a sustituir (Ctes de GeoRasterFile)
	 */
	protected void setRGBLine(BufferedImage image, int startX, int startY, int w, int h, int[] rgbArray, 
			 int offset, int scansize, int flags) {
		int [] line = new int[rgbArray.length]; 
		image.getRGB(startX, startY, w, h, line, offset, scansize);
		if (flags == GeoRasterFile.RED_BAND)
			for (int i=0; i<line.length; i++)
				line[i] = (line[i] & 0x0000ffff) | (rgbArray[i] & 0xffff0000);
		else if (flags == GeoRasterFile.GREEN_BAND)
			for (int i=0; i<line.length; i++)
				line[i] = (line[i] & 0x00ff00ff) | (rgbArray[i] & 0xff00ff00);
		else if (flags == GeoRasterFile.BLUE_BAND)
			for (int i=0; i<line.length; i++)
				line[i] = (line[i] & 0x00ffff00) | (rgbArray[i] & 0xff0000ff);
		image.setRGB(startX, startY, w, h, line, offset, scansize);
	}
	
	/**
	 * Asigna al objeto Image la mezcla entre los valores que ya tiene y los valores 
	 * con los dato de la imagen contenidos en el vector de enteros. De los valores RGB
	 * que ya contiene se mantienen las bandas que no coinciden con el valor de flags. La
	 * banda correspondiente a flags es sustituida por los datos del vector.
	 * @param image	imagen con los datos actuales
	 * @param startX	inicio de la posicin en X dentro de la imagen
	 * @param startY	inicio de la posicin en X dentro de la imagen
	 * @param w	Ancho de la imagen
	 * @param h	Alto de la imagen
	 * @param rgbArray	vector que contiene la banda que se va a sustituir
	 * @param offset	desplazamiento
	 * @param scansize	tamao de imagen recorrida por cada paso
	 * @param origBand	Banda origen del GeoRasterFile
	 * @param destBandFlag	banda que se va a sustituir (Ctes de GeoRasterFile)
	 */
	protected void setRGBLine(BufferedImage image, int startX, int startY, int w, int h, int[] rgbArray, 
			 int offset, int scansize, int origBand, int destBandFlag) {
		int [] line = new int[rgbArray.length]; 
		image.getRGB(startX, startY, w, h, line, offset, scansize);
		if (origBand == 0 && destBandFlag == GeoRasterFile.RED_BAND)
			for (int i=0; i<line.length; i++)
				line[i] = (line[i] & 0x0000ffff) | (rgbArray[i] & 0xffff0000);
		else if (origBand == 1 && destBandFlag == GeoRasterFile.GREEN_BAND)
			for (int i=0; i<line.length; i++)
				line[i] = (line[i] & 0x00ff00ff) | (rgbArray[i] & 0xff00ff00);
		else if (origBand == 2 && destBandFlag == GeoRasterFile.BLUE_BAND)
			for (int i=0; i<line.length; i++)
				line[i] = (line[i] & 0x00ffff00) | (rgbArray[i] & 0xff0000ff);
		
		else if (origBand == 0 && destBandFlag == GeoRasterFile.GREEN_BAND)
			for (int i=0; i<line.length; i++)
				line[i] = (line[i] & 0xffff00ff) | ((rgbArray[i] & 0x00ff0000) >> 8) ;
		else if (origBand == 0 && destBandFlag == GeoRasterFile.BLUE_BAND)
			for (int i=0; i<line.length; i++)
				line[i] = (line[i] & 0xffffff00) | ((rgbArray[i] & 0x00ff0000) >> 16);
		else if (origBand == 1 && destBandFlag == GeoRasterFile.RED_BAND)
			for (int i=0; i<line.length; i++)
				line[i] = (line[i] & 0xff00ffff) | ((rgbArray[i] & 0x0000ff00) << 8);
		
		else if (origBand == 1 && destBandFlag == GeoRasterFile.BLUE_BAND)
			for (int i=0; i<line.length; i++)
				line[i] = (line[i] & 0xffffff00) | ((rgbArray[i] & 0x0000ff00) >> 8);
		else if (origBand == 2 && destBandFlag == GeoRasterFile.RED_BAND)
			for (int i=0; i<line.length; i++)
				line[i] = (line[i] & 0xff00ffff) | ((rgbArray[i] & 0x000000ff) << 16);
		else if (origBand == 2 && destBandFlag == GeoRasterFile.GREEN_BAND)
			for (int i=0; i<line.length; i++)
				line[i] = (line[i] & 0xffff00ff) | ((rgbArray[i] & 0x000000ff) << 8);
		image.setRGB(startX, startY, w, h, line, offset, scansize);
	}
	
	private void showOnOpen() {
  		// Report en la apertura (quitar)
  		System.out.println("Fichero GDAL '"+getName()+"' abierto.");
  		System.out.println("Version = "+file.version);
  		System.out.println("   Size = ("+file.width+","+file.height+")");
  		try {
			System.out.println("   NumBands = ("+file.getRasterCount()+")");
		} catch (GdalException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
  		//file.pintaInfo();
  		file.pintaPaleta();

	}

	/* (non-Javadoc)
	 * @see org.cresques.io.GeoRasterFile#updateImage(int, int, org.cresques.cts.ICoordTrans, java.awt.Image, int, int)
	 */
	public Image updateImage(int width, int height, ICoordTrans rp, Image img, int origBand, int destBandFlag)throws SupersamplingNotSupportedException{
		int line, pRGBArray[] = null;
			
		if(mustVerifySize()){
			// Work out the correct aspect for the setView call.
			double dFileAspect = (double)v.width()/(double)v.height();
			double dWindowAspect = (double)width /(double)height;
	
			if (dFileAspect > dWindowAspect) {
			  height =(int)((double)width/dFileAspect);
			} else {
			  width = (int)((double)height*dFileAspect);
			}
		}
		
		// Set the view
		boolean[] orientation = getOrientation();
		file.setView(v.minX(), v.maxY(), v.maxX(), v.minY(),
			width, height, orientation);
		
		if(width<=0)width=1;
		if(height<=0)height=1;
		
		pRGBArray = new int[width];
		try {
			setBand(RED_BAND,   rBandNr);
			setBand(GREEN_BAND, gBandNr);
			setBand(BLUE_BAND,  bBandNr);
			file.setAlpha(getAlpha());
			if(img!=null){
				if(orientation[1]){
					for (line=0; line < height; line++) {
						file.readLineRGBA(pRGBArray);
						setRGBLine((BufferedImage) img, 0, height - 1 - line, width, 1, pRGBArray, 0, width, origBand, destBandFlag);
					}	
				}else{
					for (line=0; line < height; line++) {
						file.readLineRGBA(pRGBArray);
						setRGBLine((BufferedImage) img, 0, line, width, 1, pRGBArray, 0, width, origBand, destBandFlag);
					}
				}
				return img;
			}else{
				Image image = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
				if(orientation[1]){
					for (line=0; line < height; line++) {
						file.readLineRGBA(pRGBArray);
						setRGBLine((BufferedImage) image, 0, height - 1 - line, width, 1, pRGBArray, 0, width);
					}	
				}else{
					for (line=0; line < height; line++) {
						file.readLineRGBA(pRGBArray);
						setRGBLine((BufferedImage) image, 0, line, width, 1, pRGBArray, 0, width);
					}
				}
				return image;
			}
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		return img;
	}
	
	/* (non-Javadoc)
	 * @see org.cresques.io.GeoRasterFile#getData(int, int, int)
	 */
	public Object getData(int x, int y, int band) {
		if(file != null){
			Object[] data = file.getData(x, y);
			return data[band];
		}
		return null;
	}
	
	/**
	 * Devuelve los datos de una ventana solicitada
	 * @param ulX	coordenada X superior izda.
	 * @param ulY	coordenada Y superior derecha.
	 * @param sizeX	tamao en X de la ventana.
	 * @param sizeY tamao en Y de la ventana.
	 * @param band	Banda solicitada.
	 */
	public byte[] getWindow(int ulX, int ulY, int sizeX, int sizeY, int band){
		
		return null;
	}
	
	/**
	 * Obtiene la zona (Norte / Sur)
	 * @return true si la zona es norte y false si es sur
	 */
	
	public boolean getZone(){
		
		return false;
	}
	
	/**
	 *Devuelve el nmero de zona UTM
	 *@return Nmero de zona 
	 */
	
	public int getUTM(){
		
		return 0;	
	}
	
	/**
	 * Obtiene el sistema de coordenadas geograficas
	 * @return Sistema de coordenadas geogrficas
	 */
	public String getGeogCS(){
		
		return new String("");	
	}
	
	/**
	 * Devuelve el tamao de bloque
	 * @return Tamao de bloque
	 */
	public int getBlockSize(){
	  return file.getBlockSize();
	}
		
	/**
	 * Obtiene el objeto que contiene los metadatos
	 */
	public Metadata getMetadata() {
		if(file != null)
			return file.getMetadataJavaObject();
		else 
			return null;
	}
	
	/**
	 * Obtiene el flag que dice si la imagen est o no georreferenciada
	 * @return true si est georreferenciada y false si no lo est.
	 */
	public boolean isGeoreferenced() {
		return file.isGeoreferenced();
	}
    
	/**
	 * Obtiene los parmetros de la transformacin afn que corresponde con los elementos de
	 * un fichero tfw.
	 * <UL> 
	 * <LI>[1]tamao de pixel en X</LI>
	 * <LI>[2]rotacin en X</LI>
	 * <LI>[4]rotacin en Y</LI>
	 * <LI>[5]tamao de pixel en Y</LI>
	 * <LI>[0]origen en X</LI>
	 * <LI>[3]origen en Y</LI>
	 * </UL>
	 * Este mtodo debe ser reimplementado por el driver si tiene esta informacin. En principio
	 * Gdal es capaz de proporcionarla de esta forma.
	 * 
	 * En caso de que exista fichero .rmf asociado al raster pasaremos de la informacin de georreferenciacin
	 * del .tfw y devolveremos la que est asociada al rmf
	 * @return vector de double con los elementos de la transformacin afn.
	 */
	public double[] getTransform(){
		if(file != null && file.trans != null && !this.rmfExists())
			return file.trans.adfgeotransform;
		else{
			if(this.rmfExists){
				double[] rmfGeoref = {	rmfTransform.getTranslateX(), 
										rmfTransform.getScaleX(),
										rmfTransform.getShearX(), 
										rmfTransform.getTranslateY(),
										rmfTransform.getShearY(),
										rmfTransform.getScaleY()};
				return rmfGeoref;
			}
			return null;
		}
		
	}
}


