package com.hardcode.gdbms.engine.strategies;

import com.hardcode.driverManager.DriverLoadException;
import com.hardcode.gdbms.engine.data.driver.DriverException;
import com.hardcode.gdbms.engine.instruction.CustomAdapter;
import com.hardcode.gdbms.engine.instruction.EvaluationException;
import com.hardcode.gdbms.engine.instruction.SelectAdapter;
import com.hardcode.gdbms.engine.instruction.SemanticException;
import com.hardcode.gdbms.engine.instruction.UnionAdapter;
import com.hardcode.gdbms.parser.ParseException;

import java.io.IOException;


/**
 * Interfaz que define las operaciones que se pueden realizar con los
 * DataSource. Las distintas implementaciones de esta interfaz sern las
 * encargadas del uso de los indices, del algoritmo usado para cada operacin,
 * ...
 */
public abstract class Strategy {
    /**
     * Realiza una select a partir de la instruccin que se pasa como parmetro
     *
     * @param instr Objeto con la informacin sobre las tablas que entran en
     *        juego en la instruccin, campos, expresiones condicionales, ...
     *
     * @return DataSource con el resultado de la instruccion
     *
     * @throws DriverException Excepcin que indica error del driver accediendo
     *         a  los datos
     * @throws SemanticException Si se produce algn error semntico
     * @throws IOException Si se produce un error accediendo a las estructuras
     *         de datos internas
     * @throws EvaluationException If the evaluation of any expression fails
     * @throws RuntimeException bug
     */
    public OperationDataSource select(SelectAdapter instr)
        throws DriverException, SemanticException, IOException, 
            EvaluationException {
        throw new RuntimeException(
            "This strategy does not support select execution");
    }

    /**
     * Realiza una union a partir de la instruccin que se pasa como parmetro
     *
     * @param instr Objeto con la informacin sobre las tablas que entran en
     *        juego en la instruccin
     *
     * @return DataSource con el resultado de la instruccion
     *
     * @throws DriverException Excepcin que indica error del driver accediendo
     *         a  los datos
     * @throws IOException Si se produce un erro accediendo a los datos
     * @throws SemanticException Si se produce algn error semntico
     * @throws EvaluationException If there's any problem during expresion evaluation
     * @throws ParseException If the union statement contains a select statement and its parse fails
     * @throws DriverLoadException If cannot find the suitable driver to access the data
     * @throws RuntimeException bug
     */
    public OperationDataSource union(UnionAdapter instr)
        throws DriverException, IOException, SemanticException, DriverLoadException, ParseException, EvaluationException {
        throw new RuntimeException(
            "This strategy does not support union execution");
    }

    /**
     * Makes a custom query
     *
     * @param instr The instruction specifying the custom query
     *
     * @return The result DataSource
     *
     * @throws SemanticException If there is some semantic error in the
     *         expression
     * @throws RuntimeException bug
     */
    public OperationDataSource custom(CustomAdapter instr)
        throws SemanticException {
        throw new RuntimeException(
            "This strategy does not support custom queries execution");
    }
}
