package org.agil.core.spatialindex;

//import com.soluzionasf.onis.datastructs.onisfile.*;
//import com.soluzionasf.onis.geometria.*;
//import com.soluzionasf.onis.util.Log;

import com.vividsolutions.jts.geom.Envelope;
import java.io.*;
import java.util.*;

/**
 *  <p>
 *
 *  <b>Descripcion:</b> Realiza el calculo de los nodos que intersectan con el
 *  HyperCube pasado al su constructor, liberando al objeto RTree de dicho
 *  calculo.
 *
 */
public class RTreeIterator implements Iterator {
	private List nodos;
	private int num_nodo = 0;
	// n del nodo por el que va el iterator
	private Thread th = null;


	/**
	 *  Construye un nuevo RTreeIterator, calculando los nodos que intersectan
	 *  con el Rect
	 *
	 *@param  r     rectangulo solicitado
	 *@param  file  fichero del RTree
	 */
	public RTreeIterator(Envelope r, PageFile file) {
		//   Log.msg(hashCode()+"RTreeIterator() ->"+file.tree.getFileName());
		nodos = Collections.synchronizedList(new ArrayList());
		num_nodo = 0;
		HyperCube h = new HyperCube(new Point(r.getMinX(), r.getMinY()), new Point(r.getMaxX(), r.getMaxY()));
		//lanza un thread que calcula la interseccion, dejando los nodos en el List 'nodos'
		th = new Intersector(h, (AbstractNode) file.readNode(0), nodos);
		th.start();
	}


	/**
	 *@return    true si aun quedan nodos por recorrer
	 */
	public synchronized boolean hasNext() {
		// es posible que lo he haya dado tiempo a llenar el array
		while (th.isAlive()) {
			if (num_nodo < nodos.size()) {
				return true;
			}
			// dejamos que pase otro thread
			Thread.currentThread().yield();
		}
		if (num_nodo < nodos.size()) {
			return true;
		}
		return false;
	}


	/**
	 *  Siguiente objeto del arbol que cumple la condicion. Se supone que
	 *  previamente ha pasado ha por hasNext().
	 *
	 *@return    la siguiente pagina valida a leer del fichero de datos
	 */
	public synchronized Object next() {
		Object obj = null;
		if (num_nodo < nodos.size()) {
			obj = nodos.get(num_nodo++);
		}
		//System.out.println("Por aqui no deberia pasar nunca -------------------");
		//System.out.println("num_nodo="+num_nodo+"  total="+total_nodos+" size="+nodos.size());
		return obj;
	}


	/**
	 *  Funcion no implementada
	 */
	public void remove() {
		//no implementado
	}


	/**
	 *  Overrides the finalize method to dispose of system resources or to
	 *  perform other cleanup when the RTreeIterator object is garbage collected.
	 *
	 *@exception  Throwable  Description of Exception
	 */
	protected void finalize() throws Throwable {
		//si finalizamos este objeto intentamos parar el thread.
		th.interrupt();
		super.finalize();
		nodos = null;
		//   System.out.println(hashCode()+" RTreeIterator.finalize() ->" +file.tree.getFileName());
	}


	//esta clase es la que realiza el calculo de la interseccion
	/**
	 *  Description of the Class
	 *
	 *@author    Chris Seguin
	 */
	private class Intersector extends Thread {
		private HyperCube h;
		private AbstractNode root;
		private List res;


		/**
		 *  Constructor for the Intersector object
		 *
		 *@param  h     Description of Parameter
		 *@param  root  Description of Parameter
		 *@param  res   Description of Parameter
		 */
		public Intersector(HyperCube h, AbstractNode root, List res) {
			//   Log.msg(hashCode()+" Intersector() -> "+hashCode());
			this.h = h;
			this.root = root;
			this.res = res;
		}


		/**
		 *  Main processing method for the Intersector object
		 */
		public void run() {
			//   long t1=System.currentTimeMillis();
			//   Log.msg("Intersector.run() -> "+hashCode());
			intersection(h, root, res);
			//   Log.msg("FIN :: Intersector.run() -> "+hashCode()+ " time: " + (System.currentTimeMillis()-t1));
		}


		/**
		 *  Description of the Method
		 *
		 *@param  h     Description of Parameter
		 *@param  root  Description of Parameter
		 *@param  res   Description of Parameter
		 */
		private void intersection(HyperCube h, AbstractNode root, List res) {
			if (root == null) {
				return;
			}
			HyperCube cb = root.getNodeMbb();
			//System.out.println("HYPERCUBE -> " + cb.toString());
			if (cb != null && cb.intersects(h)) {
				for (int i = 0; i < root.usedSpace; i++) {
					//System.out.println("HYPERCUBE 2 -> " + root.data[i].toString() + "POS -> " + root.branches[i]);
					if (root.data[i].intersects(h)) {
						if (root.isLeaf()) {
							res.add(new Long(root.branches[i]));
						}
						else {
							intersection(h, ((Index) root).getChild(i), res);
						}
					}
				}
			}
		}
	}
	//fin de class Intersector
}
// fin de class RTreeIterator
